
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                        %%
%% HttpClient package                                                     %%
%%                                                                        %%
%%  (c) 2000 Universit catholique de Louvain. All Rights Reserved.       %%
%%  The development of this software is supported by the PIRATES project  %%
%%  at the Universit catholique de Louvain. This file is subject to the  %%
%%  general Mozart license.                                               %%
%%                                                                        %%
%%  Author: Valentin Mesaros                                              %%
%%                                                                        %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
import 
   OS
   HttpReqRep(hReqRep:HReqRep
	      uOPrms:UOPrms
	      hMeth:HMeth
	      cHead:CHead
	      entHead:EntHead
	      genHead:GenHead
	      repHead:RepHead
	      fillNil:FillNil)

export
   cgiGET : CgiGET
   cgiPOST : CgiPOST
   cgiISINDEX : CgiISINDEX
   urlGET : UrlGET
   urlHEAD : UrlHEAD
   service : Service
   
define
%% Default Input Params
   DefInPrms = inPrms(file:nil
		      tInterval:1500
		      toFile:true
		      toStrm:false
		      overwr:true
		      proxySrv:nil
		      proxyPort:nil)

   HEX = hex(0:&0 1:&1 2:&2 3:&3 4:&4
	     5:&5 6:&6 7:&7 8:&8 9:&9
	     10:&A 11:&B 12:&C 13:&D
	     14:&E 15:&F)
   
%% Returns the hexazecimal representation of an integer (-1<I<256), as two chars string 
   fun {IntToHex I} C in
      if I < 256 andthen I>=0 then
	 C = {Int.'div' I 16}
	 [HEX.C HEX.(I-C*16)]
      else nil end
   end

%% Checks whether LS is a list of pair tuples   
   fun {IsListOfTuples LS}
      fun {Ckeck Ls}
	 case Ls of nil then true
	 [] X|Xr then
	    if {Tuple.is X} andthen {Width X} == 2 then {Ckeck Xr}
	    else false end
	 end
      end
   in
      case LS of nil then false
      else {Ckeck LS}
      end
   end
    
%% Checks whether S is a substring of LS 
   fun {IsSub S LS CaseSens}
      if {List.length LS} >= {List.length S} then
	 case S of nil then true
	 [] X|Xr then
	    if CaseSens then
	       if X=={List.nth LS 1} then	    
		  {IsSub Xr {List.drop LS 1} CaseSens}
	       else false end
	    else
	       if {Char.toLower X}=={Char.toLower {List.nth LS 1}} then	    
		  {IsSub Xr {List.drop LS 1} CaseSens}
	       else false end
	    end
	 end
      else false end
   end 
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of Service class definition
   class Service
      attr
	 OHReqRep
	 inputPrms
	 httpReq
	 codeEntity: CodeEntity
	 makeEntity: MakeEntity
	 redirServ: RedirectedService
	 service: Service
	 treatErr: TreatErr
      meth AddJoin(Lr1 R1 R2)
	 case Lr1 of nil then skip
	 [] X|Xr then
	    if {IsFree R1.X} andthen {Value.hasFeature R2 X} andthen {IsDet R2.X} then
	       R1.X = R2.X
	    end
	    {self AddJoin(Xr R1 R2)}
	 end
      end
      meth bytesRead($)
	 {@OHReqRep bytesRead($)}
      end
      meth closeAll(Trace)
	 try
	    {@OHReqRep closeAll(Trace)}
	 catch E then raise E end
	 end
      end
%% Encodes the special chars within the LS string as %nn format
      meth CodeEntity(LS CE)
	 EXCEPT = [&* &_ &- &. &$ &@ &! &( &) &,]
	 fun {Code Ls ToHex}
	    case Ls of nil then nil
	    [] X|Xr then
	       if ToHex then {List.flatten {IntToHex X}|{Code Xr false}}
	       elseif {List.member X EXCEPT} orelse {Char.isAlNum X} orelse
		  X>255 then
		  X|{Code Xr false}
	       else
		  if X==129 then &%|{Code Xr true}
		  elseif X==32 then &+|{Code Xr false}
		  else &%|{List.flatten {IntToHex X}|{Code Xr false}}
		  end
	       end
	    end
	 end
      in
	 CE={Code LS false}
      end
      meth fetchRate($)
	 {@OHReqRep fetchRate($)}	 
      end
      meth getTimeInterval($)
	 {@OHReqRep getTimeInterval($)}
      end
      meth init(InPrms HttpReq)
	 inputPrms<-InPrms
	 httpReq<-HttpReq
	 OHReqRep <- {New HReqRep init}
      end
      meth Service(Meth Uri Entity ?OutPrms ?HPrms ?Location<=_)
	 Prms Host HAddr HPort IsURL % Entity Addr 
      in
	 try
	    {self InPrmsProc(@inputPrms {Record.arity DefInPrms} Prms)}
	    {self UriProc(Uri Host HAddr HPort IsURL)}
	    if IsURL then %% URL
	       {@OHReqRep urlForm(Meth Entity Host HAddr HPort Prms @httpReq OutPrms HPrms Location)}
	    else %% filePath
	       {self InitHPrms(HPrms)}
	       {FillNil {Record.toListInd HPrms}}
	       {@OHReqRep fileForm(Host Prms OutPrms)}
	       Location=nil
	    end
	 catch E then
	    {self TreatErr(Prms OutPrms HPrms)}
	    raise E end
	 end
      end
      meth HAddrProc(SAddr ?HAddr)
	 if SAddr == nil then HAddr = "/"
	 else
	    HAddr = SAddr 
	 end
      end
%% Extraxting the port      
      meth HostProc(SHost ?HPort)
	 if {List.member CHead.'Colon' SHost $} then
	    Port
	 in
	    {String.token SHost CHead.'Colon' _ Port}
	    if Port\=nil andthen {String.isInt Port} then
	       HPort = {String.toInt Port}
	    else raise errUriChrs end end
	 else
	    HPort = 80
	 end
      end
%% Initialising the HttpRepPrms record      
      meth InitHPrms(HPrms)
	 {Record.make httpRep {List.flatten
			       {Record.arity GenHead}|
			       {Record.arity RepHead}|
			       {Record.arity EntHead}|
			       [content_length]} HPrms}
      end
%% Initialising the OutPrms        
      meth InPrmsProc(InPrms DefPrms OPrms)
	 if {IsDet InPrms} then
	    case DefPrms
	    of nil then OPrms=InPrms
	    [] P|Pr then
	       if {Value.hasFeature InPrms P} andthen
		  {IsDet InPrms.P} andthen
		  InPrms.P\=nil then
		  {self InPrmsProc(InPrms Pr OPrms)}
	       else
		  {self InPrmsProc({AdjoinAt InPrms P DefInPrms.P} Pr OPrms)}
	       end
	    end
	 else
	    OPrms=DefInPrms
	 end
      end

%% Concatenates the entities      
      meth MakeEntity(InLs OutLs)
	 try
	    case InLs of nil then OutLs=nil
	    [] X|Xr then
	       if X\=nil then
		  if Xr\= nil then	        
		     {List.append {List.append {self SplitPair(X $)} &&|nil $} {self MakeEntity(Xr $)} OutLs}	       
		  else
		     {self SplitPair(X OutLs)}  
		  end
	       else OutLs=nil
	       end
	    end
	 catch E then
	    raise E end
	 end
      end

%% Get the service taking location into account      
      meth RedirectedService(Meth Uri Pairs ?OutPrms ?HttpRep ?Redirect)
	 TRedirect Location
      in
	 try
	    {self Service(Meth Uri Pairs OutPrms HttpRep Location)}
	 catch E then
	    raise E end
	 end
	 if Location\=nil then
	    {self init(@inputPrms @httpReq)}
	    try
	       {self RedirectedService(Meth Location Pairs OutPrms HttpRep TRedirect)}
	       Redirect=[Location]|TRedirect
	    catch E then
	       raise E end
	    end
	 else
	    Redirect = nil
	 end
      end

      meth setTimeInterval(TInterval)
	 {@OHReqRep setTimeInterval(TInterval)}
      end

%% Splits the tuple pair and insert '=' between elements      
      meth SplitPair(Tpl ?Pair)
	 Name Value
      in
	 if {Tuple.is Tpl} andthen {Width Tpl}==2 then
	    Name=Tpl.1
	    Value=Tpl.2
	 else raise errCgiPrms end
	 end
	 Pair ={self CodeEntity(Name $)}|CHead.'EQ'|{self CodeEntity(Value $)}
      end
      meth startTemp
	 {@OHReqRep startTemp}
      end
      meth stopTemp
	 {@OHReqRep stopTemp}	
      end

%% Makes sure the output params are filled whether an error occurs      
      meth TreatErr(InPrms OutPrms HPrms)
	 if {IsFree OutPrms} then
	    {Record.make outPrms UOPrms OutPrms}
	    {self AddJoin(UOPrms OutPrms InPrms)}
	 end
	 if {IsFree HPrms} then
	    {self InitHPrms(HPrms)}
	 end
	 OutPrms.time={OS.time}*1000 
	 {FillNil {Record.toListInd HPrms}}
	 {FillNil {Record.toListInd OutPrms}}
      end

%% Given the Uri, extracts Host Addr Port and specifies if it is a URL or FilePath      
      meth UriProc(Uri ?Host ?HAddr ?HPort ?IsURL)
	 HUri SAddr THost
      in
	 if {Char.isAlNum {List.nth Uri 1}}==false then
	    raise errUriChrs end
	 end
	 if {IsSub CHead.'FileForm' Uri false} then
	    Host = {List.drop Uri {List.length CHead.'FileForm'}}
	    HAddr=nil
	    HPort=nil
	    IsURL=false
	 else
	    if {IsSub CHead.'HttpForm' Uri false} then
	       HUri = {List.drop Uri {List.length CHead.'HttpForm'}}
	    else
	       HUri = Uri
	    end
	    {String.token HUri CHead.'SLH' THost {List.append [CHead.'SLH'] $ SAddr}}
	    try
	       {self HostProc(THost HPort)}
	    catch E then raise E end end
	    {self HAddrProc(SAddr HAddr)}
	    if {List.member &: THost} andthen HPort==80 then
	       {String.token THost &: Host _}
	    else
	       Host = THost
	    end
	    IsURL=true
	 end
      end
   end
% END of Service class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of CgiGET class definition
   class CgiGET from Service
      meth init(InPrms HttpReq)
	 (Service, init(InPrms HttpReq))
      end
      meth getService(Uri Pairs ?OutPrms ?HPrms)
	 RServ=@redirServ
	 TErr=@treatErr
	 MEntity=@makeEntity
	 Redirect
	 HAddr
      in
	 if {IsFree Pairs} orelse Pairs==nil then HAddr=Uri
	 elseif {IsListOfTuples Pairs}==false then
	    {self TErr(@inputPrms OutPrms HPrms)}
	    raise errCgiPrms end
	 else 
	    HAddr = {List.flatten Uri|CHead.'QS'|{self MEntity(Pairs $)}}
	    if {List.length HAddr} > 1024 then
	       {self TErr(@inputPrms OutPrms HPrms)}
	       raise errLargeCgiPrms end
	    end
	 end
	 try
	    {self RServ(HMeth.get HAddr nil
			?OutPrms ?HPrms ?Redirect)}
	    if {IsFree OutPrms.redirect} then 
	       OutPrms.redirect=Redirect
	    end
	 catch E then
	    raise E end
	 end
      end
   end
% END of CgiGET class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of CgiPOST class definition
   class CgiPOST from Service
      meth init(InPrms HttpReq)
	 (Service, init(InPrms HttpReq))
      end
      meth getService(Uri Pairs ?OutPrms ?HPrms)
	 Serv=@service
	 TErr=@treatErr
	 MEntity=@makeEntity
	 Entity
      in
	 if {IsFree Pairs} orelse Pairs==nil then Entity = nil
	 elseif {IsListOfTuples Pairs} then
	    Entity = {List.flatten {self MEntity(Pairs $)}}
	 elseif {List.is Pairs} then
	    Entity = Pairs
	 else
	    {self TErr(@inputPrms OutPrms HPrms)}
	    raise errCgiPrms end
	 end
	 try
	    {self Serv(HMeth.post Uri Entity ?OutPrms ?HPrms)}
	    if {IsFree OutPrms.redirect} then 
	       OutPrms.redirect=nil
	    end
	 catch E then
	    raise E end
	 end
      end
   end
% END of CgiPOST class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of CgiISINDEX class definition
   class CgiISINDEX from Service
      meth init(InPrms HttpReq)
	 (Service, init(InPrms HttpReq))
      end
      meth getService(Uri Pairs ?OutPrms ?HPrms)
	 RServ=@redirServ
	 TErr=@treatErr
	 CEntity=@codeEntity
	 Redirect
	 HAddr
      in
	 if {IsFree Pairs} orelse Pairs==nil then HAddr=Uri
	 elseif {String.is Pairs}==false then
	    {self TErr(@inputPrms OutPrms HPrms)}
	    raise errCgiPrms end
	 else
	    HAddr = {List.flatten Uri|CHead.'QS'|{self CEntity(Pairs $)}}
	    if {List.length HAddr} > 1024 then
	       {self TErr(@inputPrms OutPrms HPrms)}
	       raise errLargeCgiPrms end
	    end
	 end
	 try
	    {self RServ(HMeth.get HAddr nil ?OutPrms ?HPrms ?Redirect)}
	    if {IsFree OutPrms.redirect} then 
	       OutPrms.redirect=Redirect
	    end
	 catch E then
	    raise E end
	 end	 
      end
   end
% END of CgiISINDEX class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of UrlGET class definition
   class UrlGET from Service      
      meth init(InPrms HttpReq)
	 (Service, init(InPrms HttpReq))
      end
      meth getService(Uri ?OutPrms ?HttpRep)
	 RServ=@redirServ
	 Redirect
      in
	 try
	    {self RServ(HMeth.get Uri nil ?OutPrms ?HttpRep ?Redirect)}
	    OutPrms.redirect=Redirect
	 catch E then
	    raise E end
	 end
      end      
   end
% END of UrlGET class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of UrlHEAD class definition
   class UrlHEAD from Service
      meth init(InPrms HttpReq)
	 (Service, init(InPrms HttpReq))
      end
      meth getService(Uri ?OutPrms ?HttpRep)
	 RServ=@redirServ
	 Redirect
      in
	 try
	    {self RServ(HMeth.head Uri nil ?OutPrms ?HttpRep ?Redirect)}
	    if {IsFree OutPrms.redirect} then 
	       OutPrms.redirect=Redirect
	    end
	 catch E then
	    raise E end
	 end
      end
   end
% END of UrlHEAD class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
end% end of the functor







