
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                        %%
%% HttpClient package                                                     %%
%%                                                                        %%
%%  (c) 2000 Universit catholique de Louvain. All Rights Reserved.       %%
%%  The development of this software is supported by the PIRATES project  %%
%%  at the Universit catholique de Louvain. This file is subject to the  %%
%%  general Mozart license.                                               %%
%%                                                                        %%
%%  Author: Valentin Mesaros                                              %%
%%                                                                        %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor 
import 
   OS
   OpenConnection(oConnection:OConnection)
export
   hReqRep: HReqRep
   uOPrms: UOPrms
   hMeth:HMeth
   cHead:CHead
   entHead:EntHead
   genHead:GenHead
   repHead:RepHead
   fillNil:FillNil
define
%  User Agent IDs
   AgID = agID(hVers:"1.0")
   
%  User Output Params
   UOPrms = [file rateStrm sOut sizeRead status time redirect]

%  Header Methods   
   HMeth = hMeth(get:"GET" post:"POST" head:"HEAD")
   
%  Header Chars   
   CHead = cHead('LF':0x0A
		 'CR':0x0D
		 'SP':0x20
		 'HTTP':"HTTP/"
		 'CRLF':"\r\n"
		 'HttpForm':"http://"
		 'FileForm':"file://"
		 'BSLH':0x5C
		 'SLH':0x2F
		 'AND':0x26
		 'EQ':0x3D
		 'QS':0x3F
		 'Comma':0x2C
		 'Tilda':0x7E
		 'Colon':0x3A
		 'Dot':0x2E)

%  No Content Status Codes   
   NoContStat = ["300" "301" "302" "304"]   

%  Entity Headers
   EntHead = entHead(allow:"Allow:"
		     content_language:"Content-Language:"
		     content_encoding:"Content-Encoding:"
		     content_type:"Content-Type:"
		     expires:"Expires:"
		     last_modified:"Last-Modified:"
		     link:"Link:"
		     title:"Title:"
		     uri:"URI:")
   %content_language:"Content-Length:" % explicit added

%  General Headers
   GenHead = genHead(date:"Date:"
		     pragma:"Pragma:")
   
%  Request Headers
   ReqHead = reqHead(accept:"Accept:"
		     accept_charset:"Accept-Charset:"
		     accept_encoding:"Accept-Encoding:"
		     accept_language:"Accept-Language:"
		     authorization:"Authorization:"
		     'from':"From:"
		     if_modified_since:"If-Modified-Since:"
		     refer:"Refer:"
		     user_agent:"User-Agent:")
   
%  Response Headers
   RepHead = repHead(location:"Location:"
		     retry_after:"Retry-After:"
		     server:"Server:"
		     www_authenticate:"WWW-Authenticate:")
   
%% Returns a new list with the surplus Blank chars droped out from the given list LS
   fun {DropCTL LS}
      case LS
      of nil then nil
      [] X|Xr then
	 if X > 0x20 then    
	    X|{DropCTL Xr}
	 elseif X == CHead.'SP' then
	    case Xr
	    of nil then nil
	    [] S|_ then
	       if S == CHead.'SP' then
		  {DropCTL Xr}
	       else
		  X|{DropCTL Xr}
	       end
	    end
	 else {DropCTL Xr}   % CTL chars
	 end	    
      end 
   end

%% Returns a new list with Chr chars droped out from the given list LS   
   fun {DropChr LS Chr}
      case LS of nil then nil
      [] X|Xr then
	 if X \= Chr then X|{DropChr Xr Chr}
	 else {DropChr Xr Chr} end
      end
   end
   
%% Bounds to nil the Y element of the LS list of X#Y pairs, if is unbound    
   proc {FillNil LS}
      case LS of nil then skip
      [] X|Xr then
	 %{System.show X.2}
	 if {IsFree X.2} andthen {Value.isFuture X.2}==false then X.2 = nil end
	 %if {IsFree X.2} then  X.2 = nil end
	 {FillNil Xr}
      end
   end
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of HReqRep class definition
   class HReqRep
      attr
	 HSocket
	 FileName
      meth AddrFix(Host InAddr OutAddr InitFileName)
	 Split FileN 
      in
	 {List.last {String.tokens InAddr CHead.'SLH'} Split}
	 case Split of nil then
	    {List.last {String.tokens Host CHead.'SLH'} FileN}
	    OutAddr = InAddr
	 else
	    if {List.nth Split 1}\=CHead.'Tilda' then
	       FileN = Split
	       OutAddr = InAddr
	    else
	       if {List.last InAddr} \= CHead.'SLH' then
		  OutAddr={List.append InAddr [CHead.'SLH']}
	       else OutAddr = InAddr end
	       FileN = {DropChr Split CHead.'Tilda'}	       
	    end
	 end
	 if {IsDet InitFileName} andthen InitFileName\=nil then
	    FileName<-InitFileName
	 else	    
	    FileName<-{String.token FileN CHead.'QS' $ _}
	 end
      end
      meth bytesRead($)
	 {@HSocket bytesRead($)}
      end
      meth closeAll(Trace)
	 try
	    {@HSocket closeAll}
	    if Trace==false then
	       {@HSocket remFile}
	    end
	 catch errNoFile then skip
	 [] errNoConnection then skip
	 [] errRemFile then skip
	 [] E then raise E end
	 end
      end
      meth FileFix(FilePath FName)
	 if {List.member &/ FilePath} then
	    {List.last {String.tokens FilePath &/} FName}
	 elseif {List.member &\\ FilePath} then
	    {List.last {String.tokens FilePath &\\} FName}
	 else
	    FName = FilePath
	 end
      end
      meth fileForm(FilePath InPrms ?OutPrms)
	 {self InitOutPrms(InPrms OutPrms)}
	 if {IsDet InPrms.file} andthen InPrms.file\=nil then
	    FileName<-InPrms.file
	 else
	    FileName<-{self FileFix(FilePath $)}
	 end
	 try
	    {@HSocket openConnection(tInterval:InPrms.tInterval
				     host:FilePath rateStrm:OutPrms.rateStrm
				     isURL:false)}
	    {self FetchEntity(nil nil InPrms OutPrms OutPrms.sizeRead)}
	 catch E then
	    raise E end
	 finally
	    try
	       {@HSocket closeConnection}
	    catch _ then skip
	    end
	 end 
	 OutPrms.time = {OS.time}*1000
	 {FillNil {Record.toListInd OutPrms}}
      end
      meth getTimeInterval($)
	 {@HSocket getTimeInterval($)}
      end      
      meth urlForm(Method Entity HHost HAddr HPort InPrms HttpReq ?DocPrms ?HPrms ?Location)
	 fun {HeadMatch LS Rec}
	    case LS of nil then nil
	    [] X|Xr then
	       OutVS
	    in
	       if {IsDet X.2} andthen X.2\=nil then
		  if {Value.hasFeature Rec X.1} then
		     OutVS=Rec.(X.1)#" "#X.2#CHead.'CRLF'
		  else OutVS = nil
		  end
		  OutVS#{HeadMatch Xr Rec}
	       else {HeadMatch Xr Rec}
	       end
	    end
	 end
	 VSRequest
	 RateStrm
	 OAddr
	 UrlAddr
      in
	 {self AddrFix(HHost HAddr OAddr InPrms.file)}
	 if {IsFree HttpReq} then HttpReq=nil end
	 %VSRequest= Method#" "#"http://"#HHost#HAddr#CHead.'CRLF'
	 VSRequest = 
	 Method#" "#
	 UrlAddr#" "#CHead.'HTTP'#
	 AgID.hVers#CHead.'CRLF'#
	 "Host:"#" "#HHost#CHead.'CRLF'#
	 {HeadMatch {Record.toListInd HttpReq} ReqHead}#
	 {HeadMatch {Record.toListInd HttpReq} GenHead}
	 try
	    if InPrms.proxySrv\=nil andthen InPrms.proxyPort\=nil then
	       {self OpenConnection(InPrms.tInterval InPrms.proxySrv InPrms.proxyPort RateStrm)}
	       UrlAddr = CHead.'HttpForm'#HHost#OAddr
	    else
	       {self OpenConnection(InPrms.tInterval HHost HPort RateStrm)}
	        UrlAddr = OAddr
	    end
	    if {IsDet Entity} andthen Entity\=nil then %POST
	       {self SendRequest(VSRequest#
				 {HeadMatch {Record.toListInd HttpReq} EntHead}#
				 "Content-Length:"#" "#{List.length Entity}#
				 CHead.'CRLF'#
				 CHead.'CRLF'#Entity _)}
	    else
	       {self SendRequest(VSRequest#CHead.'CRLF' _)}
	    end
	    {self FetchResponse(Method InPrms RateStrm DocPrms HPrms Location)}
	 catch E then
	    raise E end
	 end
      end
      meth FetchEntity(LSIn SizeIn RIn ROut ?SizeOut)
	 AppendEnd LSRead
      in
	 try
	    if RIn.toFile then
	       {@HSocket openFile(@FileName [create write] RIn.overwr)}
	    end
	    if SizeIn\=nil then
	       {@HSocket readSocketEnd(ROut.sOut RIn.toFile RIn.toStrm nil SizeIn SizeOut)}
	       AppendEnd=unit
	    else  
	       thread {List.append LSIn LSRead ROut.sOut} AppendEnd=unit end
	       {@HSocket readSocketEnd(LSRead RIn.toFile RIn.toStrm nil all SizeOut)}
	    end
	 catch E then
	    AppendEnd=unit
	    raise E end
	 finally
	    if RIn.toFile then
	       try
		  {@HSocket closeFile}
	       catch _ then skip end
	    end
	    {Wait AppendEnd}
	 end
      end
      meth FetchHeaders(?ROut)
	 Header in
	 try
	    Header={DropCTL {@HSocket readSocketString($)}}
	    if Header\= nil then
	       {self HeaderProc(Header ROut)}
	       {self FetchHeaders(ROut)}
	    end
	 catch E then
	    raise E end
	 end
      end
      meth FetchResponse(Method RIn RateStrm ?ROut ?HPrms ?Location)
	 TROut THPrms
      in
	 {self InitOutPrms(RIn TROut)}
	 {self InitHPrms(THPrms)}
	 TROut.rateStrm=RateStrm
	 try
	    LHead SizeRead
	 in
	    {@HSocket readSocketBytes(LHead nil 5 _)}
	    if {self WhichHttp(LHead $)}\= nil then
	       TROut.status={List.nth {String.tokens {DropCTL {@HSocket readSocketString($)}} CHead.'SP'} 2 $}
	       {self FetchHeaders(THPrms)}
	       if {List.member TROut.status NoContStat} orelse
		  Method == HMeth.head then
		  %% redirection to other location
		  if {IsDet THPrms.location} andthen
		     Method \= HMeth.post then
		     Location = THPrms.location
		  else
		     Location=nil
		     SizeRead=0
		     ROut=TROut
		     HPrms=THPrms		     
		     ROut.sOut=nil
		     {FillNil {Record.toListInd THPrms}}
		  end
	       else
		  Size
	       in
		  ROut=TROut
		  HPrms=THPrms
		  {FillNil {Record.toListInd HPrms}}
		  if {Value.hasFeature HPrms content_length}
		     andthen HPrms.content_length\=nil then
		     Size = {String.toInt HPrms.content_length}
		  else
		     Size = nil
		  end
		  {self FetchEntity(nil Size RIn ROut SizeRead)}
		  Location = nil
	       end 
	    else
	       ROut=TROut
	       ROut.status=nil
	       HPrms=THPrms
	       {FillNil {Record.toListInd HPrms}}
	       {self FetchEntity(LHead nil RIn ROut SizeRead)}
	       Location = nil
	    end
	    TROut.sizeRead=SizeRead
	    TROut.file=@FileName
	    TROut.time = {OS.time}*1000
	 catch errNoConnection then
	    TROut.time = {OS.time}*1000
	    ROut=TROut
	    HPrms=THPrms
	    {FillNil {Record.toListInd HPrms}}
	    {FillNil {Record.toListInd ROut}}
	    Location = nil
	 [] E then raise E end 
	 finally
	    try
	       {@HSocket closeConnection}
	    catch _ then skip
	    end
	 end
      end
      meth HeaderProc(Header ?HPrms)
	 SHead SValue
	 proc {HeadMatch LS}
	    case LS of nil then skip
	    [] X|Xr then
	       if X.2 == SHead andthen {IsFree {Value.'.' HPrms X.1}} then
		  {Value.'.' HPrms X.1} = SValue
	       end
	       {HeadMatch Xr}
	    end
	 end
      in
	 {String.token Header CHead.'SP' SHead SValue}
	 {HeadMatch {List.flatten
		     {Record.toListInd GenHead}|
		     {Record.toListInd RepHead}|
		     {Record.toListInd EntHead}|
		     [content_length#"Content-Length:"]}}
      end
      meth init
	 HSocket <- {New OConnection init}
      end
      meth InitOutPrms(InPrms DocPrms)	 
	 proc {MatchFeat Ls Rin Rout}
	    case Ls of nil then skip
	    [] X|Xr then
	       if {Value.hasFeature Rin X} andthen {IsDet Rin.X}
		  andthen Rin.X\=nil then
		  Rout.X = Rin.X
	       end
	       {MatchFeat Xr Rin Rout}
	    end
	 end
	 TempPrms
      in
	 {Record.make outPrms UOPrms TempPrms}
	 {MatchFeat UOPrms InPrms TempPrms}
	 DocPrms = TempPrms
      end
      meth InitHPrms(HPrms)
	 {Record.make httpRep {List.flatten
			       {Record.arity GenHead}|
			       {Record.arity RepHead}|
			       {Record.arity EntHead}|
			       [content_length]} HPrms}
      end
      meth fetchRate($)
	 {@HSocket fetchRate($)}
      end
      meth OpenConnection(TInterval Host HPort RateStrm)
	 THost
      in
	 if {List.member CHead.'Colon' Host $} then
	    {String.token Host CHead.'Colon' THost _}
	 else
	    THost = Host
	 end
	 try
	    {@HSocket openConnection(tInterval:TInterval host:THost port:HPort
				     rateStrm:RateStrm)}
	 catch E then
	    raise E end
	 end
      end
      meth setTimeInterval(TInterval)
	 {@HSocket setTimeInterval(TInterval)}
      end
      meth startTemp
	 {@HSocket startTemp}
      end
      meth stopTemp
	 {@HSocket stopTemp}	
      end
      meth SendRequest(VSRequest LenIn)
	 try 
	    {@HSocket writeSoket(VSRequest LenIn)}
	 catch errNoConnection then skip
	 [] E then
	    raise E end
	 end
      end
      meth WhichHttp(LS V)
	 if LS==CHead.'HTTP' then V="1.x"
	 else V=nil
	 end
      end
   end   
% END of HReqRep class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end% end of the functor


