
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                        %%
%% HttpClient package                                                     %%
%%                                                                        %%
%%  (c) 2000 Universit catholique de Louvain. All Rights Reserved.       %%
%%  The development of this software is supported by the PIRATES project  %%
%%  at the Universit catholique de Louvain. This file is subject to the  %%
%%  general Mozart license.                                               %%
%%                                                                        %%
%%  Author: Valentin Mesaros                                              %%
%%                                                                        %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
import  
   OS 
   Open 
   Clock
 
export
   oConnection:OConnection
define

%% Returns the sum of list's values
   fun {LexLength LS}
      case LS of nil then 0
      [] X|Xr then
	 X+{LexLength Xr}
      end
   end

%% Checks whether S is a substring of LS 
   fun {IsSub S LS CaseSens}
      if {List.length LS} >= {List.length S} then
	 case S of nil then true
	 [] X|Xr then
	    if CaseSens then
	       if X=={List.nth LS 1} then	    
		  {IsSub Xr {List.drop LS 1} CaseSens}
	       else false end
	    else
	       if {Char.toLower X}=={Char.toLower {List.nth LS 1}} then	    
		  {IsSub Xr {List.drop LS 1} CaseSens}
	       else false end
	    end
	 end
      else false end
   end   
   
   WriteFileLock = {NewLock}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START of OConnection class definition
   class OConnection
      attr
	 oCSock: nil
	 oFile: nil
	 fPath: nil
	 TimeInterval
	 BytesCounter:0
	 RBuffer: nil
	 LastRates : nil
	 PrdBytes : 0.0
	 WSizeBuffer: 8192
	 RSizeAll: 8192
	 RSizeBuffer: 1024
	 SBlock:unit
	 FBlock:unit
	 RBlock:unit
	 ProdPort
	 FilePatern:nil
	 Overwr
	 
      meth bytesRead($)
	 @BytesCounter
      end
      meth ChoseTempFile(LS)
	 fun {Exists FileList}
	    case FileList of nil then false
	    [] X|Xr then
	       if {IsSub @FilePatern X true}
		  andthen {List.drop X {List.length @FilePatern}}==nil then
		  true
	       else {Exists Xr}
	       end
	    end
	 end
	 proc {WhichTemp F}
	    if {IsSub @FilePatern F true} then
	       NrStr
	    in
	       NrStr={List.drop F {List.length @FilePatern}}
	       if {String.isInt NrStr} then
		  TempStr={List.append @FilePatern
			      {Int.toString {String.toInt NrStr}+1}}
	       in
		  if {LexLength TempStr}>{LexLength @fPath} then
		     fPath<-TempStr
		  end
	       end
	    else skip
	    end	       
	 end
      in
	 if {Exists LS} andthen @Overwr==false then
	    FilePatern<-{List.append @FilePatern "_"}
	    {List.forAll LS proc{$ F} {WhichTemp F} end}
	    if @fPath==nil then
	       fPath<-{List.append @FilePatern "1"}
	    end
	 else fPath<-@FilePatern end	    
      end
      meth closeAll
	 Excpt
	 ConEnd
	 FileEnd
      in
	 thread
	    try
	       {self closeConnection}
	    catch E then
	       if {IsFree Excpt} then Excpt=E end
	    finally
	       ConEnd=unit
	    end 
	 end
	 thread
	    try
	       {self closeFile}
	    catch E then
	       if {IsFree Excpt} then Excpt=E end
	    finally
	       FileEnd=ConEnd
	    end 
	 end
	 {Wait FileEnd}
	 if {IsDet Excpt} then
	    raise Excpt end
	 end
      end
      meth closeConnection
	 if {IsDet @SBlock} then raise errNoConnection end end
	 if {IsDet @ProdPort} then
	    {Port.send @ProdPort nil}
	    ProdPort<-_
	 end
	 try
	    @SBlock = unit
	    {@oCSock close}
	    {self RateInit}
	    RBuffer<-nil
	    oCSock <- nil
	 catch _ then
	    raise errCloseCon end
	 end      
      end
      meth closeFile
	 if {IsDet @FBlock} then raise errNoFile end end
	 try	 
	    @FBlock = unit
	    {@oFile close}
	    oFile <- nil
	 catch _ then
	    raise errCloseFile end
	 end
      end
      meth fetchRate($)
	 CTime
      in
	 CTime = {Clock.getTime}
	 if {IsDet @RBlock} andthen @LastRates\=nil
	    andthen {IsFree @SBlock} then
	    RateDif
	 in
	    RateDif=CTime-{List.last @LastRates}.2
	    if RateDif\=0 then
	       if (CTime-{List.nth @LastRates 1}.2) =< @TimeInterval then
		  (@PrdBytes-{List.last @LastRates}.1)/({Int.toFloat RateDif})#CTime
	       else
		  0.0#CTime
	       end
	    else	       
	       (@PrdBytes-{List.last @LastRates}.1)/{Int.toFloat @TimeInterval}#CTime
	    end
	 else
	    0.0#CTime
	 end
      end
      meth getTimeInterval($)
	 @TimeInterval
      end
      meth init
	 {Clock.setClock {OS.time}*1000}
	 {Clock.startClock 10}
      end
      meth isFileDet($)
	 if @oFile \= nil then true
	 else false end
      end
      meth openConnection(tInterval:TInterval
			  host:Host rateStrm:RateStrm
			  isURL:IsURL<=true
			  port:TcpPort<=80 type:Type<=stream
			  time:Time<=10000 protocol:Protocol<="tcp")
	 TimeInterval <- TInterval
	 try
	    if IsURL then
	       oCSock <- {New  Open.socket init(type: Type 
						protocol: Protocol
						time: Time)}
	       {@oCSock connect(host:Host port:TcpPort)}
	    else
	       oCSock <- {New Open.file init(name:Host flags: [read write])}
	    end
	 catch _ then
	    raise errOpenCon end
	 end
	 ProdPort <- {Port.new RateStrm}
	 SBlock <- _	 
      end
      meth openFile(FPath Flags Owr)
	 Overwr<-Owr
	 lock WriteFileLock then
	    try	    
	       FilePatern<-FPath
	       {self ChoseTempFile({OS.getDir {OS.getCWD}})}
	       oFile <- {New Open.file init(name:@fPath
					    flags: Flags
					    mode: mode(owner:[read write] all:[read]))}
	       FBlock <- _
	    catch _ then
	       raise errOpenFile end
	    end
	 end %% end WriteFileLock
      end
      meth RateInit
	 LastRates <- [0.0#{Clock.getTime}]
	 PrdBytes <- 0.0
      end
      meth ReadSocketEnd(list:?LS tail:T<=nil len:?Len<=_)
	 LSR LenR LenIn 
      in
	 try 
	    {Wait @RBlock}
	    {@oCSock read(list:LS tail:LSR size:@RSizeAll len:LenIn)}
	    BytesCounter <- @BytesCounter+LenIn
	    {self SetRate(LenIn {Clock.getTime})}
	    if LenIn\=0 then
	       {self ReadSocketEnd(list:LSR tail:T len:LenR)}
	    else
	       LSR = T
	       Len = LenIn
	       LenR = 0
	    end	   
	 catch _ then
	    LS = nil
	    LenR = 0
	    LenIn = 0
	    raise errReadSocket end
	 finally Len = LenIn+LenR 
	 end
      end
      meth ReadSocketNrBytes(list:?LS tail:T<= nil size:S<=1024 len:?Len<=_)
	 LSR LenR LenIn
      in	
	 try    
	    {Wait @RBlock}
	    {@oCSock read(list:LS tail:LSR size:S len:LenIn)}
	    %BytesCounter <- @BytesCounter+LenIn
	    {self SetRate(LenIn {Clock.getTime})}
	    if LenIn > 0 andthen S > LenIn then
	       {self ReadSocketNrBytes(list:LSR tail:T size:(S-LenIn) len:LenR)}
	       Len = LenIn+LenR
	    else
	       LSR = T
	       Len = LenIn
	       LenR = 0
	    end
	 catch _ then
	    LS = nil
	    LenR = 0
	    LenIn = 0
	    raise errReadSocket end
	 finally Len = LenIn+LenR
	 end
      end
      meth readSocketEnd(?LS ToFile ToStrm T Size ?Len)
	 LOut LsRead
	 RLenBuffer 
	 AppendEnd WriteEnd LenAddEnd REnd LenR ThisThread
      in
	 if {IsDet @SBlock} then raise errNoConnection end end
	 RLenBuffer={List.length @RBuffer $}
	 ThisThread={Thread.this}
	 BytesCounter<-RLenBuffer
	 try 
	    if ToStrm then LS = LOut
	    else LS = nil end 
	    thread {List.append @RBuffer LsRead LOut} AppendEnd=WriteEnd end
	    if ToFile then
	       thread
		  try
		     {self WriteBlock(LOut REnd)}
		  catch E then
		     if {IsFree @FBlock} then {Thread.injectException ThisThread E} end
		  finally WriteEnd=unit 
		  end
	       end
	    else WriteEnd=unit end 
	    {self RateInit}
	    thread Len = LenR+RLenBuffer LenAddEnd=AppendEnd end
	    {self ReadSocketEnd(list:LsRead tail:T len:LenR)}
	    case Size of all then skip
	    else 
	       if Size>Len then
		  raise conPremClosed end
	       end
	    end
	 catch E then
	    if {IsFree @SBlock} then raise E end end
	 finally
	    REnd = unit
	    {Wait LenAddEnd}
	    RBuffer <- nil
	 end
      end
      meth readSocketBytes(?LS T Size ?Len)
	 LsRead
	 AppendEnd
	 RLenBuffer 
      in
	 if {IsDet @SBlock} then raise errNoConnection end end
	 RLenBuffer={List.length @RBuffer $}
	 try
	    if Size =< RLenBuffer then
	       RBuffer <- {List.takeDrop @RBuffer Size LS $}
	    else 
	       {self RateInit}
	       if Size =< @RSizeBuffer then TempBuffer in
		  thread {List.append @RBuffer {List.takeDrop LsRead Size-RLenBuffer $ TempBuffer} LS} AppendEnd=unit end
		  {self ReadSocketNrBytes(list:LsRead tail:T size:@RSizeBuffer len:Len)}
		  {Wait AppendEnd}
		  RBuffer <- TempBuffer
	       else		  
		  thread {List.append @RBuffer LsRead LS} AppendEnd=unit end
		  {self ReadSocketNrBytes(list:LsRead tail:T size:Size-RLenBuffer len:Len)}
		  {Wait AppendEnd}
		  RBuffer <- nil
	       end
	    end
	    if Size>Len+RLenBuffer then
	       raise conPremClosed end
	    end 
	 catch E then
	    if {IsFree @SBlock} then raise E end end
	 end
      end
      meth readSocketString(?S)
	 TempBuffer
      in
	 if {IsDet @SBlock} then raise errNoConnection end end
	 try
	    {self SearchForLine(@RBuffer TempBuffer S)}
	    RBuffer <- TempBuffer
	 catch E then
	    raise E end
	 end
      end
      meth remFile
	 try
	    if @fPath\=nil then 
	       {OS.unlink @fPath}
	       fPath<-nil
	    end
	 catch _ then raise errRemFile end 
	 end
      end
      meth SearchForLine(LsIn ?LsOut ?S)
	 case LsIn of nil then LS in
	    try
	       {self RateInit}
	       {self ReadSocketNrBytes(list:LS tail:nil size:@RSizeBuffer len:_)}
	       {self SearchForLine(LS LsOut S)}
	    catch E then
	       LsOut = nil   
	       S = nil
	       if {IsFree @SBlock} then raise E end end
	    end
	 [] X|Xr then
	    case X of 0x0A then            
	       LsOut = Xr   
	       S = nil
	    else
	       {List.append [X] {self SearchForLine(Xr LsOut $)} S}
	    end
	 end	 
      end
      meth SendRate
	 if {IsFree @SBlock} then 
	    {Port.send @ProdPort {self fetchRate($)}}
	 end
      end      
      meth SetRate(Bytes RTime)
	 proc {Update}
	    if {List.nth @LastRates 1}.2-{List.last @LastRates}.2 >= @TimeInterval
	       andthen {List.length @LastRates}>2
	    then
	       PrdBytes <- @PrdBytes-{List.last @LastRates}.1
	       LastRates<-{List.take @LastRates {List.length @LastRates}-1}
	       {Update}
	    end
	 end
      in	 
	 PrdBytes <- @PrdBytes+{Int.toFloat Bytes}
	 LastRates<-{List.append [{Int.toFloat Bytes}#RTime] @LastRates}
	 {Update}
	 {self SendRate}
      end
      meth setTimeInterval(TInterval)
	 TimeInterval<-TInterval
      end
      meth startTemp
	 @RBlock = unit
      end
      meth stopTemp
	 RBlock <- _
	 {self RateInit}
      end
      meth WriteBlock(LSIn REnd)
	 proc {PartRead Q Ol Or N}
	    case Q of nil then Ol=nil Or=Q
	    [] X|Xr then
	       if N =< @WSizeBuffer then 
		  {List.append [X] {PartRead Xr $ Or N+1} Ol}	 
	       else
		  Ol=nil Or=Q
	       end
	    end   
	 end
	 Oleft
	 Oright
      in
	 try
	    {PartRead LSIn Oleft Oright 1}
	    if {IsDet REnd} then
	       {@oFile write(vs:LSIn)}		  
	    elseif {IsFree REnd} then
	       {@oFile write(vs:Oleft)}	
	       {self WriteBlock(Oright REnd)}
	    end
	 catch E then
	    raise E end
	 end
      end
      meth writeSoket(VS ?Len)
	 VsLen = {VirtualString.length VS $}
	 LenIn LenW
      in
	 if {IsDet @SBlock} then raise errNoConnection end end
	 try
	    {Wait @RBlock}
	    {@oCSock write(vs:VS len:LenW)}
	    if VsLen>LenW then
	       {self writeSoket({List.drop VS VsLen $} LenIn)}
	       Len = LenW+LenIn
	    else Len = LenW end
	 catch _ then
	    Len = 0
	    raise errWriteSocket end
	 end
      end      
   end
% END of OConnection class definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end% end of the functor





















